﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Classification {
  [Item("ClassificationProblemData", "Represents an item containing all data defining a classification problem.")]
  [StorableClass]
  public class ClassificationProblemData : DataAnalysisProblemData {
    private const int MaximumClasses = 20;
    private const string ClassNamesParameterName = "ClassNames";
    private const string MisclassificationMatrixParameterName = "MisClassificationMatrix";

    public StringArray ClassNames {
      get { return ClassNamesParameter.Value; }
      protected set { ClassNamesParameter.Value = value; }
    }
    public IValueParameter<StringArray> ClassNamesParameter {
      get { return (IValueParameter<StringArray>)Parameters[ClassNamesParameterName]; }
    }

    public DoubleMatrix MisclassificationMatrix {
      get { return MisclassificationMatrixParameter.Value; }
      protected set { MisclassificationMatrixParameter.Value = value; }
    }
    public IValueParameter<DoubleMatrix> MisclassificationMatrixParameter {
      get { return (IValueParameter<DoubleMatrix>)Parameters[MisclassificationMatrixParameterName]; }
    }

    [Storable]
    private List<double> sortedClassValues;
    public IEnumerable<double> SortedClassValues {
      get { return sortedClassValues; }
    }
    public int NumberOfClasses {
      get { return sortedClassValues.Count; }
    }

    public ClassificationProblemData()
      : base() {
      Parameters.Add(new ValueParameter<StringArray>(ClassNamesParameterName, "An array of the names for all class values."));
      Parameters.Add(new ValueParameter<DoubleMatrix>(MisclassificationMatrixParameterName, "A matrix that describles the penalties for misclassifaction between the single classes."));
      sortedClassValues = new List<double>();

      RegisterParameterEvents();
      UpdateClassValues();
    }

    [StorableConstructor]
    protected ClassificationProblemData(bool deserializing) : base(deserializing) { }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserializationHook() {
      RegisterParameterEvents();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      ClassificationProblemData clone = (ClassificationProblemData)base.Clone(cloner);
      clone.RegisterParameterEvents();
      clone.UpdateClassValues();
      return clone;
    }

    protected override void OnProblemDataChanged(System.EventArgs e) {
      if (!suppressEvents) {
        UpdateClassValues();
        base.OnProblemDataChanged(e);
      }
    }

    public override void ImportFromFile(string fileName) {
      var csvFileParser = new CsvFileParser();
      csvFileParser.Parse(fileName);
      suppressEvents = true;
      Name = "Data imported from " + Path.GetFileName(fileName);
      Dataset = new Dataset(csvFileParser.VariableNames, csvFileParser.Values);
      Dataset.Name = Path.GetFileName(fileName);
      var variableNames = Dataset.VariableNames.Select(x => new StringValue(x).AsReadOnly()).ToList();
      var validTargetVariables = variableNames.Select(variable => new { Variable = variable, DistinctValues = Dataset.GetVariableValues(variable.Value, 0, 50).Distinct().Count() })
         .OrderBy(x => x.DistinctValues).Where(x => x.DistinctValues <= MaximumClasses).Select(x => x.Variable);
      if (!validTargetVariables.Any())
        throw new ArgumentException("Import of classification problem data was not successfull, because no target variable was found." +
          " A target varialbe must have at most " + MaximumClasses + " distinct values to be applicable to classification.");

      ((ConstrainedValueParameter<StringValue>)TargetVariableParameter).ValidValues.Clear();
      foreach (var variableName in validTargetVariables)
        ((ConstrainedValueParameter<StringValue>)TargetVariableParameter).ValidValues.Add(variableName);
      TargetVariable = validTargetVariables.FirstOrDefault();

      InputVariables = new CheckedItemList<StringValue>(variableNames).AsReadOnly();
      InputVariables.SetItemCheckedState(validTargetVariables.First(), false);
      int middle = (int)(csvFileParser.Rows * 0.5);
      TrainingSamplesStart = new IntValue(0);
      TrainingSamplesEnd = new IntValue(middle);
      TestSamplesStart = new IntValue(middle);
      TestSamplesEnd = new IntValue(csvFileParser.Rows);
      suppressEvents = false;
      OnProblemDataChanged(EventArgs.Empty);
    }

    private void UpdateClassValues() {
      sortedClassValues = Dataset.GetVariableValues(TargetVariable.Value).Distinct().ToList();
      sortedClassValues.Sort();
      ResetMisclassificationMatrix();
      UpdateClassNames();
    }

    private void UpdateClassNames() {
      if (ClassNames != null) DeregisterParameterValueEvents();
      StringArray array = new StringArray(NumberOfClasses);
      int i = 0;
      foreach (double classValue in SortedClassValues) {
        array[i] = "Class " + classValue;
        i++;
      }
      ClassNames = array;
      RegisterParameterValueEvents();
    }

    private void RegisterParameterEvents() {
      ClassNamesParameter.ValueChanged += new EventHandler(ClassNamesChanged);
    }
    private void RegisterParameterValueEvents() {
      ClassNames.ItemChanged += new EventHandler<EventArgs<int>>(ClassNamesChanged);
    }
    private void DeregisterParameterValueEvents() {
      ClassNames.ItemChanged -= new EventHandler<EventArgs<int>>(ClassNamesChanged);
    }

    private void ClassNamesChanged(object sender, EventArgs e) {
      UpdateMisclassifciationMatrixHeaders();
    }

    private void ResetMisclassificationMatrix() {
      double[,] matrix = new double[NumberOfClasses, NumberOfClasses];
      for (int i = 0; i < NumberOfClasses; i++) {
        for (int j = 0; j < NumberOfClasses; j++)
          if (i != j) matrix[i, j] = 1;
      }

      if (MisclassificationMatrix == null)
        MisclassificationMatrix = new DoubleMatrix(matrix);
      if (MisclassificationMatrix.Rows != NumberOfClasses)
        MisclassificationMatrix = new DoubleMatrix(matrix);
      if (MisclassificationMatrix.Columns != NumberOfClasses)
        MisclassificationMatrix = new DoubleMatrix(matrix);
    }

    private void UpdateMisclassifciationMatrixHeaders() {
      MisclassificationMatrix.RowNames = ClassNames;
      MisclassificationMatrix.ColumnNames = ClassNames;
    }


  }
}
